################################################################################
batchpick<-function(folder,ppm,peakwidth,snthresh,prefilter,mzCenterFun,integrate,mzdiff,fitgauss){

  # check inputs & load data ###################################################
  # if(length(list.dirs(folder))<1){stop("folderin: no such folder exists!")} # only works for R version >=15.
  files<-list.files(folder,recursive=FALSE);
  if(length(files)==0){stop("No files in that folder!")}
  peaks<-list(0);
  #if(any(files=="peaklists")){stop("Subfolder peaklist already contains picked peaks - start new workfolder/project or delete subfolders!")}
  #for(i in 1:length(files)){
  #     if(length(strsplit(files[i],split=".",fixed=TRUE)[[1]])<2){stop(paste("unknown format: ",files[i],sep=""))}
  #     if(strsplit(files[i],split=".",fixed=TRUE)[[1]][2]!="mzXML"){stop(paste("unknown format: ",files[i],sep=""))}
  #}
  # run centwave & save ########################################################
  cat("\n");
  cat("------------------------------------------------------------------------");
  cat("\n");
  #dir.create(paste(folder,"peaklists\\",sep=""))
  for(i in 1:length(files)){

    # load raw #################################################################
    xraw<-xcmsRaw(
      filename=paste(folder,files[i],sep=""),
      profstep=0,
      includeMSn=FALSE
    )
    # run centWave #############################################################
    getpeak<-findPeaks.centWave(
      xraw,
      ppm=ppm,
      peakwidth=peakwidth,
      snthresh=snthresh,
      prefilter=prefilter,
      mzCenterFun=mzCenterFun,
      integrate=integrate,
      mzdiff=mzdiff,
      fitgauss=fitgauss,
      #scanrange,
      #noise,
      #sleep=10,
      verbose.columns=TRUE
    )
    getpeak<-as.data.frame(getpeak[,c(1,7,4)])
    int<-order(getpeak[,2],decreasing=TRUE);
    names(getpeak)<-c("m/z","Int","RT")
    assign(strsplit(paste(files[i],sep=""),".m")[[1]][1],getpeak)
    peaks[[i]]<-getpeak
    rm(getpeak);
    # save #####################################################################
    #save(getpeak,file=paste(folder,"peaklists\\",strsplit(files[i],split=".",fixed=TRUE)[[1]][1],sep=""));
    cat("\n");
    cat(paste(i," of ",length(files)," done. ------------------------------------------",sep=""));
    cat("\n");
    ############################################################################

  }
  ##############################################################################
  return(peaks);

}
################################################################################

################################################################################
recall<-function(peaklist,mz,tolmz=10,ppm=TRUE,ret,tolret,what="mass",one=TRUE,knot=5,plotit=TRUE){

  ##############################################################################
  # search for concerned peaks #################################################
  #if(any(objects()=="search_peak")==FALSE){stop("R object / function search_peak not function! Abort.")};
  if(what!="mass" && what!="ret"){stop("what what?")};
  peaks<-search_peak(peaklist,mz,tolmz,ppm,ret,tolret);
  # collect concurring peaks ###################################################
  getit1<-c();  # expected mz
  getit2<-c();  # observed peaklist
  for(i in 1:length(peaks)){
    if(peaks[i]!="FALSE"){
      put<-as.numeric(strsplit(peaks[i],"/")[[1]]);
      if(what=="mass"){
        if(one==TRUE & length(put)==1){
          getit1<-c(getit1,rep(mz[i],length(put)));  # expected
          getit2<-c(getit2,peaklist[put,1]);         # observed
        }else{
          getit1<-c(getit1,rep(mz[i],length(put)));  # expected
          getit2<-c(getit2,peaklist[put,1]);         # observed
        }
      }else{ # what=="ret"
        if(one==TRUE & length(put)==1){
          getit1<-c(getit1,rep(ret[i],length(put))); # expected
          getit2<-c(getit2,peaklist[put,3]);         # observed
        }else{
          getit1<-c(getit1,rep(ret[i],length(put))); # expected
          getit2<-c(getit2,peaklist[put,3]);         # observed
        }
      }
    }
  }
  getit3<-c(getit1-getit2); # observed-expected
  ##############################################################################
  # stop if too few data found! ################################################
  if(length(getit3)<15){
    return("Too few data points for fit!");
  }
  ##############################################################################
  # train gam model ############################################################
  that<-data.frame(getit2,getit3)
  names(that)<-c("obs","delta")
  #attach(that)
  model<-gam(delta~s(obs,bs="ts",k=knot),data=that);
  if(plotit==TRUE){
    if(what=="mass"){
      plot(getit2,getit3,pch=19,cex=0.5,xlab="m/z",ylab="Expected - observed");
      abline(h=0,col="red");
      points(getit2[order(getit2)],predict(model)[order(getit2)],col="red",type="l",lwd=2);
      #points(getit2,getit3-predict(model),col="green",pch=19,cex=0.7);
    }else{
      plot(getit2,getit3,pch=19,cex=0.5,xlab="Retention time",ylab="Expected - observed");
      abline(h=0,col="red");
      points(getit2[order(getit2)],predict(model)[order(getit2)],col="red",type="l",lwd=2);
      #points(getit2,getit3-predict(model),col="green",pch=19,cex=0.7);
    }
  }
  ##############################################################################
  # predict -> recalibrate #####################################################
  if(what=="mass"){
    that<-data.frame("obs"=peaklist[,1],"delta"=peaklist[,1]);
    pred2<-predict.gam(model,newdata=that);
    newpeaks<-peaklist;
    newpeaks[,1]<-c(peaklist[,1]+pred2);
  }else{
    that<-data.frame("obs"=peaklist[,3],"delta"=peaklist[,3]);
    pred2<-predict.gam(model,newdata=that);
    newpeaks<-peaklist;
    newpeaks[,3]<-c(peaklist[,3]+pred2);
  }
  return(newpeaks)
  ##############################################################################
}
################################################################################

################################################################################
# convert isotopologue pattern to profile ######################################
prof<-function(pat,adduct="M+H",adduct_list,res=5,ppm=TRUE,sdfactor=1,stickfactor=4,
              getprof="centroid",cutoff=0.001){

    ############################################################################
    x<-pat[,c(1,2)];
    leng<-length(x[1,]);
    # calculate adduct mass:
    x[,1]<-x[,1]*as.numeric((adduct_list[adduct_list[,1]==adduct,4]))+as.numeric((adduct_list[adduct_list[,1]==adduct,5]));
    # rescale
    x[,2]<-c(x[,2]/x[1,2]);
    results<-x[,c(1,2)];
    results<-data.frame(results,rep(NA,length(results[,1])),rep(NA,length(results[,1])),rep(NA,length(results[,1])));
    names(results)[3]<-"upper";
    names(results)[4]<-"lower";
    names(results)[5]<-"cluster";
    ############################################################################
    # (A) calculate lower and upper thresholds for each peak and its gaussian distribution
    if(ppm==FALSE){ # based on mmu #############################################
  		for(i in 1:length(results[,1])){
  			sdRes<-as.numeric(as.numeric(res) / (1000 * 2 * sqrt( 2 * log(2) )) * as.numeric(sdfactor));
  			results[i,3]<-qnorm(as.numeric(0.001), mean = as.numeric(results[i,1]), sd = as.numeric(sdRes), lower.tail = TRUE, log.p = FALSE);
  			results[i,4]<-qnorm(as.numeric(0.999), mean = as.numeric(results[i,1]), sd = as.numeric(sdRes), lower.tail = TRUE, log.p = FALSE);
  		};
      rm(sdRes);
  	};
  	if(ppm==TRUE){ # based on ppm	##############################################
  		for(i in 1:length(results[,1])){
  			ResValue2<-as.numeric(as.numeric(res) * as.numeric(results[i,1]) / 1E6);
  			sdRes<-as.numeric(ResValue2 / ( 2 * sqrt( 2 * log(2) )) * as.numeric(sdfactor));
  			results[i,3]<-qnorm(as.numeric(0.001), mean = as.numeric(results[i,1]), sd = as.numeric(sdRes), lower.tail = TRUE, log.p = FALSE);
  			results[i,4]<-qnorm(as.numeric(0.999), mean = as.numeric(results[i,1]), sd = as.numeric(sdRes), lower.tail = TRUE, log.p = FALSE);
  		};
      rm(ResValue2, sdRes);
  	};
		############################################################################
    # group data based on quantile bounds  #####################################
    group<-as.numeric(1);
    results[1,5]<-as.numeric(group);
    for(i in 2:length(results[,1])){
    	if(as.numeric(results[i,3])<as.numeric(results[i-1,4])){
    		results[i,5]<-as.numeric(group);
    	}else{
    		group<-as.numeric(group+1);
    		results[i,5]<-as.numeric(group);
    	}
    }
    ############################################################################
    # (B) define stick discretization ##########################################
  	if(ppm==FALSE){	# dm = based on mmu	########################################
  		sticks<-c();
  		group<-as.numeric(seq(1,max(as.numeric(results[,5])),by=1));
  		egroup<-c();
  		for(i in 1:as.numeric(max(results[,5]))){
  			low<-as.numeric(min(as.numeric(results[results[,5]==group[i],3])));
  			up<-as.numeric(max(as.numeric(results[results[,5]==group[i],4])));
  			getstick<-seq(  from=as.numeric(low),  to=as.numeric(up),  by=(as.numeric(res/stickfactor)/1000) );
  			sticks<-c(sticks,getstick);
  			egroup<-c(egroup,rep(group[i],length(getstick)));
  		}
  	}
  	if(ppm==TRUE){  # dm = based on ppm ########################################
  		sticks<-c();
  		count<-c(0);
  		group<-as.numeric(seq(1,max(as.numeric(results[,5])),by=1));
  		egroup<-c();
  		for(i in 1:max(as.numeric(results[,5]))){
  			sticks<-c(sticks,as.numeric(min(as.numeric(results[results[,5]==group[i],3]))));
  			egroup<-c(egroup,group[i]);
  			up<-as.numeric(max(as.numeric(results[results[,5]==group[i],4])));
  			count<-as.numeric(count+1);
  			while(as.numeric(sticks[count])<as.numeric(up)){
  				maxstick<-as.numeric(sticks[count]);
  				sticks<-c(sticks,maxstick+as.numeric(as.numeric(res/stickfactor)*maxstick/1e6));
  				count<-as.numeric(count+1);
  				egroup<-c(egroup,group[i]);
  			} # while-loop
  		} # for-loop
  	}
    ############################################################################
    # (C) calculate f(stick[j]) = summe( N(peak[i],function(mass)) ) ###########
    if(ppm==FALSE){		# based on mmu ###########################################
    		profiles<-rep(0,length(sticks));	        # iteratively stores all profile data of all single peaks in result data set
    		for(i in 1:length(results[,1])){        # per peak in result data set
    			group<-as.numeric(results[i,5]);		  # define group to which the concerned peak belongs
    			subprofile<-sticks[egroup==group];    # subset of m/z stick values for which calculation should be conducted!
    			collect<-rep(0,length(subprofile));	  # store profile results over subprofile m/z for each peak
    			leng<-length(subprofile);						  # number of data points in the cluster to which that peak belongs
    			sdRes<-as.numeric(as.numeric(res) / (1000 * 2 * sqrt(2*log(2) )) * as.numeric(sdfactor));
    			meanval<-as.numeric(as.numeric(results[i,2])/dnorm(results[i,1],mean=results[i,1],sd=sdRes,log=FALSE));
    			for(j in 1:leng){				              # calculate profile data over subprofile d(m/z) values
    			   collect[j]<-as.numeric(dnorm(subprofile[j],mean=results[i,1],sd=sdRes,log=FALSE)*meanval);
    			}
    			profiles[egroup==group]<-as.numeric(profiles[egroup==group]+collect);
    		} # for-loop i: per peak
	  } # ppm==FALSE
  	if(ppm==TRUE){ # based on ppm, description see above
  		profiles<-rep(0,length(sticks));
  		for(i in 1:length(results[,1])){
  			group<-as.numeric(results[i,5]);
  			subprofile<-sticks[egroup==group];
  			collect<-rep(0,length(subprofile));
  			leng<-length(subprofile);
  			ResValue<-as.numeric(as.numeric(res)*as.numeric(results[i,1])/1E6)
  			sdRes<-as.numeric(as.numeric(ResValue) / (2*sqrt(2*log(2) )) * as.numeric(sdfactor));
  			meanval<-as.numeric(as.numeric(results[i,2])/dnorm(results[i,1],mean=results[i,1],sd=sdRes,log=FALSE));
  			for(j in 1:leng){
  			   collect[j]<-as.numeric(dnorm(subprofile[j],mean=results[i,1],sd=sdRes,log=FALSE)*meanval);
  			}
  			profiles[egroup==group]<-as.numeric(profiles[egroup==group]+collect);
  		} # for-loop i: per peak
  	} # ppm==TRUE
    if(getprof=="sticks"){
      results<-data.frame(sticks,profiles)
      names(results)<-c("stick m/z","abundance")
    }
    ############################################################################
    # (D) calculate valleys ####################################################
    if(getprof=="intensoid" || getprof=="centroid"){
      valleymz<-data.frame(c(sticks[1]),c(profiles[1]));
      names(valleymz)[1]<-"m/z";
      names(valleymz)[2]<-"intensity";
      for(i in 2: (length(profiles)-1)){
      	if(profiles[i]<profiles[i-1] && profiles[i]<profiles[i+1]){
      	valleymz<-rbind(valleymz,c(sticks[i],profiles[i]))}
      };
    }
    ############################################################################
    # (E) calculate surviving peaks	############################################
    if(getprof=="intensoid"){
      results5<-results[1,];
      for(i in 1:length(valleymz[,1])){
      	results4<-results[as.numeric(results[,1])>as.numeric(valleymz[i,1]),];
      	if(i!=length(valleymz[,1])){results4<-results4[as.numeric(results4[,1])<as.numeric(valleymz[i+1,1]),]};
      	results4<-results4[as.numeric(results4[,2])==max(as.numeric(results4[,2])),];
      	results5<-rbind(results5,results4);
      } # for loop
      results2<-results5[-1,-c(3,4,5)];
      results<-results2
      names(results)<-c("intensoid m/z","abundance")
    }
    ############################################################################
    # (F) calculate centroid data ##############################################
    if(getprof=="centroid"){
        these2<-rep(1,length(sticks));	# (a) create vector "these2" indicating which stick belongs to which original m/z peak
        if(length(valleymz[,1])>1){
    	     for(i in 1:length(valleymz[,1])){
              these2[sticks>=valleymz[i,1]]<-i
            }
        }
        results2<-data.frame(c(0),c(0),c(0));
        names(results2)[1]<-"centroid m/z";
        names(results2)[2]<-"abundance";
        names(results2)[3]<-"original m/z";
        for(i in 1:length(levels(as.factor(these2)))){
        	centroid<-(sum(profiles[these2==i]*sticks[these2==i])/sum(profiles[these2==i]));
        	these3<-profiles[these2==i];
        	results2<-rbind(results2,c(centroid,max(these3)));	# take maximum!
        }
        results2<-results2[-1,-3];
        results2[,2]<-as.numeric(as.numeric(results2[,2]) / as.numeric(results2[1,2]));
        results<-results2;
        names(results)<-c("centroid m/z","abundance");
    }
    ############################################################################
    # cutoff
    results<-results[as.numeric(results[,2])>=as.numeric(cutoff),]
    results<-results[order(as.numeric(results[,2]),decreasing=TRUE),]
    results[,2]<-as.numeric(as.numeric(results[,2]) / as.numeric(results[1,2]));
    ############################################################################
    return(results);

}
################################################################################

################################################################################
recall<-function(peaklist,mz,tolmz=10,ppm=TRUE,ret,tolret,what="mass",one=TRUE,knot=5,plotit=TRUE){

  ##############################################################################
  # search for concerned peaks #################################################
  #if(any(objects()=="search_peak")==FALSE){stop("R object / function search_peak not function! Abort.")};
  if(what!="mass" && what!="ret"){stop("what what?")};
  peaks<-search_peak(peaklist,mz,tolmz,ppm,ret,tolret);
  # collect concurring peaks ###################################################
  getit1<-c();  # expected mz
  getit2<-c();  # observed peaklist
  for(i in 1:length(peaks)){
    if(peaks[i]!="FALSE"){
      put<-as.numeric(strsplit(peaks[i],"/")[[1]]);
      if(what=="mass"){
        if(one==TRUE & length(put)==1){
          getit1<-c(getit1,rep(mz[i],length(put)));  # expected
          getit2<-c(getit2,peaklist[put,1]);         # observed
        }else{
          getit1<-c(getit1,rep(mz[i],length(put)));  # expected
          getit2<-c(getit2,peaklist[put,1]);         # observed
        }
      }else{ # what=="ret"
        if(one==TRUE & length(put)==1){
          getit1<-c(getit1,rep(ret[i],length(put))); # expected
          getit2<-c(getit2,peaklist[put,3]);         # observed
        }else{
          getit1<-c(getit1,rep(ret[i],length(put))); # expected
          getit2<-c(getit2,peaklist[put,3]);         # observed
        }
      }
    }
  }
  getit3<-c(getit1-getit2); # observed-expected
  ##############################################################################
  # stop if too few data found! ################################################
  if(length(getit3)<15){
    return("Too few data points for fit!");
  }
  ##############################################################################
  # train gam model ############################################################
  that<-data.frame(getit2,getit3)
  names(that)<-c("obs","delta")
  #attach(that)
  model<-gam(delta~s(obs,bs="ts",k=knot),data=that);
  if(plotit==TRUE){
    if(what=="mass"){
      plot(getit2,getit3,pch=19,cex=0.5,xlab="m/z",ylab="Expected - observed");
      abline(h=0,col="red");
      points(getit2[order(getit2)],predict(model)[order(getit2)],col="red",type="l",lwd=2);
      #points(getit2,getit3-predict(model),col="green",pch=19,cex=0.7);
    }else{
      plot(getit2,getit3,pch=19,cex=0.5,xlab="Retention time",ylab="Expected - observed");
      abline(h=0,col="red");
      points(getit2[order(getit2)],predict(model)[order(getit2)],col="red",type="l",lwd=2);
      #points(getit2,getit3-predict(model),col="green",pch=19,cex=0.7);
    }
  }
  ##############################################################################
  # predict -> recalibrate #####################################################
  if(what=="mass"){
    that<-data.frame("obs"=peaklist[,1],"delta"=peaklist[,1]);
    pred2<-predict.gam(model,newdata=that);
    newpeaks<-peaklist;
    newpeaks[,1]<-c(peaklist[,1]+pred2);
  }else{
    that<-data.frame("obs"=peaklist[,3],"delta"=peaklist[,3]);
    pred2<-predict.gam(model,newdata=that);
    newpeaks<-peaklist;
    newpeaks[,3]<-c(peaklist[,3]+pred2);
  }
  return(newpeaks)
  ##############################################################################
}
################################################################################

################################################################################
# search list of peaks within list of peaks ####################################
search_peak<-function(peaklist,mz,tolmz=5,ppm=TRUE,ret,tolret){
  ##############################################################################
  # redefine order of peaklist & mz, keep ord1 and ord2 to resort
  peaks<-peaklist[order(peaklist[,1],decreasing=TRUE),];
  ord1<-order(peaklist[,1],decreasing=TRUE);
  mass<-mz[order(mz,decreasing=TRUE)];
  ord2<-order(mz,decreasing=TRUE);
  ##############################################################################
  result<-rep("FALSE",length(mz));
  k<-c(2);
  leng<-length(peaks[,1]);
  ##############################################################################
  for(i in 1:length(mass)){
    deletes<-c();
    # calculate bounds per target mass #########################################
    if(ppm==TRUE){
    	target_low<-as.numeric(mass[i])-(as.numeric(tolmz)*(as.numeric(mass[i])/1e6));
    	target_up<-as.numeric(mass[i])+(as.numeric(tolmz)*(as.numeric(mass[i])/1e6));
    }else{
    	target_low<-as.numeric(mass[i])-(as.numeric(tolmz)/1000);
    	target_up<-as.numeric(mass[i])+(as.numeric(tolmz)/1000);
    }
    # reset k ##################################################################
   	while(target_up < as.numeric(peaks[k,1])  &&  k<as.numeric(leng) ){
     k<-c(k+1)
    };
   	n<-c(k);
   	while(target_low <= as.numeric(peaks[n,1]) && n<as.numeric(leng)){
		  n<-c(n+1);
    };
   	if(n==leng){n<-c(n-1)}; # nonsense?
   	for(  f in (k-1):(n+1)  ){
    	if((as.numeric(peaks[f,1]) >= as.numeric(target_low)) &&
    		( as.numeric(peaks[f,1]) <= as.numeric(target_up))  &&
    		( as.numeric(peaks[f,3]) >= (as.numeric(ret[ord2[i]])-as.numeric(tolret[i]))) &&
    		( as.numeric(peaks[f,3]) <= (as.numeric(ret[ord2[i]])+as.numeric(tolret[i])))){
          deletes<-c(deletes,f);
        }
    }
    # save results #############################################################
    if(length(deletes)>0){
     result[ord2[i]]<-as.character(ord1[deletes[1]]);
     if(length(deletes)>1){
      for(j in 2:length(deletes)){
         result[ord2[i]]<-paste(result[ord2[i]],"/",as.character(ord1[deletes[j]]))
      }
     }
    }
	}
	return(result);
}
################################################################################

################################################################################
# subtract list2 from another list1 = find peaks of list1 not part of list2 ####
sub_peak<-function(peaklist1,peaklist2,tolmz=10,ppm=TRUE,tolret=0.5,cutoff=10){
  ##############################################################################
  getit<-c();
  ID<-seq(1:length(peaklist2[,1]));
  for(i in 1:length(peaklist1[,1])){
    if(ppm==TRUE){
    	mzmin<-as.numeric(peaklist1[i,1])-(as.numeric(tolmz)*(as.numeric(peaklist1[i,1])/1e6));
    	mzmax<-as.numeric(peaklist1[i,1])+(as.numeric(tolmz)*(as.numeric(peaklist1[i,1])/1e6));
    }else{
    	mzmin<-as.numeric(peaklist1[i,1])-(as.numeric(tolmz)/1000);
    	mzmax<-as.numeric(peaklist1[i,1])+(as.numeric(tolmz)/1000);
    }
    IDsub<-ID[ # faster: (1) first check mass... ###############################
      peaklist2[,1]>=mzmin &
      peaklist2[,1]<=mzmax]
    if(length(IDsub)==0){
      getit<-c(getit,i);
    }else{ # ... (2) and then check RT, ... ####################################
      rtmin<-as.numeric(peaklist1[i,3])-tolret;
      rtmax<-as.numeric(peaklist1[i,3])+tolret;
      IDsub<-IDsub[
        peaklist2[IDsub,3]>=rtmin &
        peaklist2[IDsub,3]<=rtmax];
      if(length(IDsub)==0){
        getit<-c(getit,i);
      }else{ # ... (3) and then - optionally - check intensity #################
        if(cutoff!=FALSE){
          IDsub<-IDsub[(peaklist2[IDsub,2]) < ( peaklist1[i,2]/cutoff) ];
          if(length(IDsub)==0){
            getit<-c(getit,i);
          }
        }
      }
    }
    ############################################################################
  }
  ##############################################################################
  return(peaklist1[getit,]);
  ##############################################################################
};
################################################################################

################################################################################

################################################################################